import os

import autofit as af
from autoastro.plotters import fit_galaxy_plotters
from autolens.plotters import phase_plotters, hyper_plotters


def setting(section, name):
    return af.conf.instance.visualize.get(section, name, bool)


def plot_setting(name):
    return setting("plots", name)


def figure_setting(name):
    return setting("figures", name)


class AbstractVisualizer:
    def __init__(self, image_path):
        self.image_path = image_path or ""
        try:
            os.makedirs(self.image_path)
        except (FileExistsError, FileNotFoundError):
            pass
        self.plot_in_kpc = af.conf.instance.visualize.get(
            "figures", "plot_in_kpc", bool
        )
        self.include_mask = figure_setting("include_mask")
        self.include_critical_curves = figure_setting("include_critical_curves")
        self.include_caustics = figure_setting("include_caustics")
        self.plot_ray_tracing_all_at_end_png = plot_setting(
            "plot_ray_tracing_all_at_end_png"
        )
        self.plot_ray_tracing_all_at_end_fits = plot_setting(
            "plot_ray_tracing_all_at_end_fits"
        )
        self.plot_ray_tracing_as_subplot = plot_setting("plot_ray_tracing_as_subplot")
        self.plot_ray_tracing_profile_image = plot_setting(
            "plot_ray_tracing_profile_image"
        )
        self.plot_ray_tracing_source_plane = plot_setting(
            "plot_ray_tracing_source_plane_image"
        )
        self.plot_ray_tracing_convergence = plot_setting("plot_ray_tracing_convergence")
        self.plot_ray_tracing_potential = plot_setting("plot_ray_tracing_potential")
        self.plot_ray_tracing_deflections = plot_setting("plot_ray_tracing_deflections")
        self.plot_ray_tracing_magnification = plot_setting(
            "plot_ray_tracing_magnification"
        )


class SubPlotVisualizer(AbstractVisualizer):
    def __init__(self, image_path):
        super().__init__(image_path)
        self.subplot_path = f"{image_path}subplots/"
        try:
            os.makedirs(self.subplot_path)
        except FileExistsError:
            pass


class PhaseGalaxyVisualizer(AbstractVisualizer):
    def __init__(self, image_path):
        super().__init__(image_path)
        self.plot_galaxy_fit_all_at_end_png = plot_setting(
            "plot_galaxy_fit_all_at_end_png"
        )
        self.plot_galaxy_fit_all_at_end_fits = plot_setting(
            "plot_galaxy_fit_all_at_end_fits"
        )
        self.plot_galaxy_fit_as_subplot = plot_setting("plot_galaxy_fit_as_subplot")
        self.plot_galaxy_fit_image = plot_setting("plot_galaxy_fit_image")
        self.plot_galaxy_fit_noise_map = plot_setting("plot_galaxy_fit_noise_map")
        self.plot_galaxy_fit_model_image = plot_setting("plot_galaxy_fit_model_image")
        self.plot_galaxy_fit_residual_map = plot_setting("plot_galaxy_fit_residual_map")
        self.plot_galaxy_fit_chi_squared_map = plot_setting(
            "plot_galaxy_fit_chi_squared_map"
        )

    def plot_galaxy_fit_subplot(self, fit, path_suffix=""):
        if self.plot_galaxy_fit_as_subplot:
            fit_galaxy_plotters.subplot(
                fit=fit,
                include_mask=self.include_mask,
                output_path=f"{self.image_path}/{path_suffix}",
                output_format="png",
            )

    def plot_fit_individuals(
        self, fit, plot_all=False, image_format="png", path_suffix=""
    ):
        if plot_all:
            plot_image = True
            plot_noise_map = True
            plot_model_image = True
            plot_residual_map = True
            plot_chi_squared_map = True
        else:
            plot_image = self.plot_galaxy_fit_image
            plot_noise_map = self.plot_galaxy_fit_noise_map
            plot_model_image = self.plot_galaxy_fit_model_image
            plot_residual_map = self.plot_galaxy_fit_residual_map
            plot_chi_squared_map = self.plot_galaxy_fit_chi_squared_map
        fit_galaxy_plotters.individuals(
            fit=fit,
            include_mask=self.include_mask,
            plot_image=plot_image,
            plot_noise_map=plot_noise_map,
            plot_model_image=plot_model_image,
            plot_residual_map=plot_residual_map,
            plot_chi_squared_map=plot_chi_squared_map,
            output_path=f"{self.image_path}/{path_suffix}",
            output_format=image_format,
        )


class PhaseDatasetVisualize(SubPlotVisualizer):
    def __init__(self, masked_dataset, image_path):
        super().__init__(image_path)
        self.masked_dataset = masked_dataset

        self.include_image_plane_pix = figure_setting(
            "include_image_plane_pixelization_grid"
        )
        self.include_positions = figure_setting("include_positions")

        self.plot_dataset_as_subplot = plot_setting("plot_dataset_as_subplot")
        self.plot_dataset_data = plot_setting("plot_dataset_data")
        self.plot_dataset_noise_map = plot_setting("plot_dataset_noise_map")
        self.plot_dataset_psf = plot_setting("plot_dataset_psf")

        self.plot_dataset_signal_to_noise_map = plot_setting(
            "plot_dataset_signal_to_noise_map"
        )
        self.plot_dataset_absolute_signal_to_noise_map = plot_setting(
            "plot_dataset_absolute_signal_to_noise_map"
        )
        self.plot_dataset_potential_chi_squared_map = plot_setting(
            "plot_dataset_potential_chi_squared_map"
        )
        self.plot_fit_all_at_end_png = plot_setting("plot_fit_all_at_end_png")
        self.plot_fit_all_at_end_fits = plot_setting("plot_fit_all_at_end_fits")
        self.plot_fit_as_subplot = plot_setting("plot_fit_as_subplot")
        self.plot_fit_of_planes_as_subplot = plot_setting(
            "plot_fit_of_planes_as_subplot"
        )
        self.plot_fit_inversion_as_subplot = plot_setting(
            "plot_fit_inversion_as_subplot"
        )
        self.plot_fit_data = plot_setting("plot_fit_data")
        self.plot_fit_noise_map = plot_setting("plot_fit_noise_map")
        self.plot_fit_signal_to_noise_map = plot_setting("plot_fit_signal_to_noise_map")
        self.plot_fit_model_data = plot_setting("plot_fit_model_data")
        self.plot_fit_residual_map = plot_setting("plot_fit_residual_map")
        self.plot_fit_normalized_residual_map = plot_setting(
            "plot_fit_normalized_residual_map"
        )
        self.plot_fit_chi_squared_map = plot_setting("plot_fit_chi_squared_map")
        self.plot_fit_contribution_maps = plot_setting("plot_fit_contribution_maps")
        self.plot_fit_inversion_residual_map = plot_setting(
            "plot_fit_inversion_residual_map"
        )
        self.plot_fit_pixelization_normalized_residuals = plot_setting(
            "plot_fit_inversion_normalized_residual_map"
        )
        self.plot_fit_inversion_chi_squared_map = plot_setting(
            "plot_fit_inversion_chi_squared_map"
        )
        self.plot_fit_inversion_regularization_weights = plot_setting(
            "plot_fit_inversion_regularization_weight_map"
        )
        self.plot_fit_subtracted_images_of_planes = plot_setting(
            "plot_fit_subtracted_images_of_planes"
        )
        self.plot_fit_model_images_of_planes = plot_setting(
            "plot_fit_model_images_of_planes"
        )
        self.plot_fit_plane_images_of_planes = plot_setting(
            "plot_fit_plane_images_of_planes"
        )
        self.plot_hyper_model_image = plot_setting("plot_hyper_model_image")
        self.plot_hyper_galaxy_images = plot_setting("plot_hyper_galaxy_images")

    def plot_ray_tracing(self, tracer, during_analysis):
        positions = self.masked_dataset.positions if self.include_positions else None
        mask = self.masked_dataset.mask if self.include_mask else None
        phase_plotters.ray_tracing_of_phase(
            tracer=tracer,
            grid=self.masked_dataset.grid,
            during_analysis=during_analysis,
            mask=mask,
            include_critical_curves=self.include_critical_curves,
            include_caustics=self.include_caustics,
            positions=positions,
            plot_in_kpc=self.plot_in_kpc,
            plot_as_subplot=self.plot_ray_tracing_as_subplot,
            plot_all_at_end_png=self.plot_ray_tracing_all_at_end_png,
            plot_all_at_end_fits=self.plot_ray_tracing_all_at_end_fits,
            plot_image=self.plot_ray_tracing_profile_image,
            plot_source_plane=self.plot_ray_tracing_source_plane,
            plot_convergence=self.plot_ray_tracing_convergence,
            plot_potential=self.plot_ray_tracing_potential,
            plot_deflections=self.plot_ray_tracing_deflections,
            visualize_path=self.image_path,
            subplot_path=self.subplot_path,
        )


class PhaseImagingVisualizer(PhaseDatasetVisualize):
    def __init__(self, masked_dataset, image_path):
        super(PhaseImagingVisualizer, self).__init__(
            masked_dataset=masked_dataset, image_path=image_path
        )

        self.plot_dataset_psf = plot_setting("plot_dataset_psf")

        self.plot_hyper_model_image = plot_setting("plot_hyper_model_image")
        self.plot_hyper_galaxy_images = plot_setting("plot_hyper_galaxy_images")

        self.plot_imaging()

    @property
    def masked_imaging(self):
        return self.masked_dataset

    def plot_imaging(self):
        mask = self.masked_dataset.mask if self.include_mask else None
        positions = self.masked_dataset.positions if self.include_positions else None

        phase_plotters.imaging_of_phase(
            imaging=self.masked_dataset.imaging,
            mask=mask,
            positions=positions,
            unit_label="arcsec",
            kpc_per_arcsec=None,
            plot_as_subplot=self.plot_dataset_as_subplot,
            plot_image=self.plot_dataset_data,
            plot_noise_map=self.plot_dataset_noise_map,
            plot_psf=self.plot_dataset_psf,
            plot_signal_to_noise_map=self.plot_dataset_signal_to_noise_map,
            plot_absolute_signal_to_noise_map=self.plot_dataset_absolute_signal_to_noise_map,
            plot_potential_chi_squared_map=self.plot_dataset_potential_chi_squared_map,
            visualize_path=self.image_path,
            subplot_path=self.subplot_path,
        )

    def plot_fit(self, fit, during_analysis):

        phase_plotters.imaging_fit_of_phase(
            fit=fit,
            during_analysis=during_analysis,
            include_mask=self.include_mask,
            include_critical_curves=self.include_critical_curves,
            include_caustics=self.include_caustics,
            include_positions=self.include_positions,
            plot_in_kpc=self.plot_in_kpc,
            include_image_plane_pix=self.include_image_plane_pix,
            plot_all_at_end_png=self.plot_fit_all_at_end_png,
            plot_all_at_end_fits=self.plot_fit_all_at_end_fits,
            plot_fit_as_subplot=self.plot_fit_as_subplot,
            plot_fit_of_planes_as_subplot=self.plot_fit_of_planes_as_subplot,
            plot_inversion_as_subplot=self.plot_fit_inversion_as_subplot,
            plot_image=self.plot_fit_data,
            plot_noise_map=self.plot_fit_noise_map,
            plot_signal_to_noise_map=self.plot_fit_signal_to_noise_map,
            plot_model_image=self.plot_fit_model_data,
            plot_residual_map=self.plot_fit_residual_map,
            plot_normalized_residual_map=self.plot_fit_normalized_residual_map,
            plot_chi_squared_map=self.plot_fit_chi_squared_map,
            plot_inversion_residual_map=self.plot_fit_inversion_residual_map,
            plot_inversion_normalized_residual_map=self.plot_fit_normalized_residual_map,
            plot_inversion_chi_squared_map=self.plot_fit_inversion_chi_squared_map,
            plot_inversion_regularization_weights=(
                self.plot_fit_inversion_regularization_weights
            ),
            plot_subtracted_images_of_planes=self.plot_fit_subtracted_images_of_planes,
            plot_model_images_of_planes=self.plot_fit_model_images_of_planes,
            plot_plane_images_of_planes=self.plot_fit_plane_images_of_planes,
            visualize_path=self.image_path,
            subplot_path=self.subplot_path,
        )

    def plot_hyper_images(self, last_results):
        mask = self.masked_dataset.mask
        if self.include_mask and mask is not None and last_results is not None:
            phase_plotters.plot_hyper_images_for_phase(
                hyper_model_image=last_results.hyper_model_image,
                hyper_galaxy_image_path_dict=last_results.hyper_galaxy_image_path_dict,
                mask=self.masked_dataset.mask,
                unit_label="arcsec",
                kpc_per_arcsec=None,
                plot_hyper_model_image=self.plot_hyper_model_image,
                plot_hyper_galaxy_images=self.plot_hyper_galaxy_images,
                visualize_path=self.image_path,
            )


class PhaseInterferometerVisualizer(PhaseDatasetVisualize):
    def __init__(self, masked_dataset, image_path):
        super(PhaseInterferometerVisualizer, self).__init__(
            masked_dataset=masked_dataset, image_path=image_path
        )

        self.plot_dataset_uv_wavelengths = plot_setting("plot_dataset_uv_wavelengths")
        self.plot_dataset_primary_beam = plot_setting("plot_dataset_primary_beam")

        self.plot_hyper_model_image = plot_setting("plot_hyper_model_image")
        self.plot_hyper_galaxy_images = plot_setting("plot_hyper_galaxy_images")

        self.plot_interferometer()

    @property
    def masked_interferometer(self):
        return self.masked_dataset

    def plot_interferometer(self):

        phase_plotters.interferometer_of_phase(
            interferometer=self.masked_interferometer.interferometer,
            unit_label="arcsec",
            kpc_per_arcsec=None,
            plot_as_subplot=self.plot_dataset_as_subplot,
            plot_visibilities=self.plot_dataset_data,
            plot_uv_wavelengths=self.plot_dataset_uv_wavelengths,
            plot_primary_beam=self.plot_dataset_primary_beam,
            visualize_path=self.image_path,
            subplot_path=self.subplot_path,
        )

    def plot_fit(self, fit, during_analysis):

        phase_plotters.interferometer_fit_of_phase(
            fit=fit,
            during_analysis=during_analysis,
            include_positions=self.include_positions,
            plot_in_kpc=self.plot_in_kpc,
            include_mask=self.include_mask,
            include_critical_curves=self.include_critical_curves,
            include_caustics=self.include_caustics,
            include_image_plane_pix=self.include_image_plane_pix,
            plot_all_at_end_png=self.plot_fit_all_at_end_png,
            plot_all_at_end_fits=self.plot_fit_all_at_end_fits,
            plot_fit_as_subplot=self.plot_fit_as_subplot,
            plot_inversion_as_subplot=self.plot_fit_inversion_as_subplot,
            plot_visibilities=self.plot_fit_data,
            plot_noise_map=self.plot_fit_noise_map,
            plot_signal_to_noise_map=self.plot_fit_signal_to_noise_map,
            plot_model_visibilities=self.plot_fit_model_data,
            plot_residual_map=self.plot_fit_residual_map,
            plot_normalized_residual_map=self.plot_fit_normalized_residual_map,
            plot_chi_squared_map=self.plot_fit_chi_squared_map,
            plot_inversion_residual_map=self.plot_fit_inversion_residual_map,
            plot_inversion_normalized_residual_map=self.plot_fit_normalized_residual_map,
            plot_inversion_chi_squared_map=self.plot_fit_inversion_chi_squared_map,
            plot_inversion_regularization_weights=(
                self.plot_fit_inversion_regularization_weights
            ),
            visualize_path=self.image_path,
            subplot_path=self.subplot_path,
        )

    def plot_hyper_images(self, last_results):
        mask = self.masked_dataset.mask
        if self.include_mask and mask is not None and last_results is not None:
            phase_plotters.plot_hyper_images_for_phase(
                hyper_model_image=last_results.hyper_model_image,
                hyper_galaxy_image_path_dict=last_results.hyper_galaxy_image_path_dict,
                mask=self.masked_interferometer.mask,
                plot_hyper_model_image=self.plot_hyper_model_image,
                plot_hyper_galaxy_images=self.plot_hyper_galaxy_images,
                visualize_path=self.image_path,
            )


class HyperGalaxyVisualizer(SubPlotVisualizer):
    def __init__(self, image_path):
        super().__init__(image_path)
        self.plot_hyper_galaxy_subplot = plot_setting("plot_hyper_galaxy_subplot")

    def hyper_galaxy_subplot(
        self,
        hyper_galaxy_image,
        contribution_map,
        noise_map,
        hyper_noise_map,
        chi_squared_map,
        hyper_chi_squared_map,
    ):
        hyper_plotters.subplot_of_hyper_galaxy(
            hyper_galaxy_image_sub=hyper_galaxy_image,
            contribution_map_sub=contribution_map,
            noise_map_sub=noise_map,
            hyper_noise_map_sub=hyper_noise_map,
            chi_squared_map_sub=chi_squared_map,
            hyper_chi_squared_map_sub=hyper_chi_squared_map,
            output_path=self.subplot_path,
            output_format="png",
        )



import autofit as af
import autoarray as aa
from autolens.plotters import ray_tracing_plotters, hyper_plotters
from autolens.plotters.fit_imaging_plotters import fit_imaging_plotters
from autolens.plotters.fit_interferometer_plotters import fit_interferometer_plotters


def imaging_of_phase(
    imaging,
    mask,
    positions,
    kpc_per_arcsec,
    unit_label,
    plot_as_subplot,
    plot_image,
    plot_noise_map,
    plot_psf,
    plot_signal_to_noise_map,
    plot_absolute_signal_to_noise_map,
    plot_potential_chi_squared_map,
    visualize_path,
    subplot_path,
):

    output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=visualize_path, folder_names=["imaging"]
    )

    if plot_as_subplot:

        aa.plot.imaging.subplot(
            imaging=imaging,
            mask=mask,
            positions=positions,
            unit_label=unit_label,
            unit_conversion_factor=kpc_per_arcsec,
            output_path=subplot_path,
            output_format="png",
        )

    aa.plot.imaging.individual(
        imaging=imaging,
        mask=mask,
        positions=positions,
        unit_label=unit_label,
        unit_conversion_factor=kpc_per_arcsec,
        plot_image=plot_image,
        plot_noise_map=plot_noise_map,
        plot_psf=plot_psf,
        plot_signal_to_noise_map=plot_signal_to_noise_map,
        plot_absolute_signal_to_noise_map=plot_absolute_signal_to_noise_map,
        plot_potential_chi_squared_map=plot_potential_chi_squared_map,
        output_path=output_path,
        output_format="png",
    )


def interferometer_of_phase(
    interferometer,
    unit_label,
    kpc_per_arcsec,
    plot_as_subplot,
    plot_visibilities,
    plot_uv_wavelengths,
    plot_primary_beam,
    visualize_path,
    subplot_path,
):

    output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=visualize_path, folder_names=["interferometer"]
    )

    if plot_as_subplot:

        aa.plot.interferometer.subplot(
            interferometer=interferometer,
            unit_label=unit_label,
            unit_conversion_factor=kpc_per_arcsec,
            output_path=subplot_path,
            output_format="png",
        )

    aa.plot.interferometer.individual(
        interferometer=interferometer,
        plot_visibilities=plot_visibilities,
        plot_u_wavelengths=plot_uv_wavelengths,
        plot_v_wavelengths=plot_uv_wavelengths,
        plot_primary_beam=plot_primary_beam,
        unit_label=unit_label,
        unit_conversion_factor=kpc_per_arcsec,
        output_path=output_path,
        output_format="png",
    )


def ray_tracing_of_phase(
    tracer,
    grid,
    during_analysis,
    mask,
    include_critical_curves,
    include_caustics,
    positions,
    plot_in_kpc,
    plot_as_subplot,
    plot_all_at_end_png,
    plot_all_at_end_fits,
    plot_image,
    plot_source_plane,
    plot_convergence,
    plot_potential,
    plot_deflections,
    visualize_path,
    subplot_path,
):
    output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=visualize_path, folder_names=["ray_tracing"]
    )

    if plot_as_subplot:

        ray_tracing_plotters.subplot(
            tracer=tracer,
            grid=grid,
            mask=mask,
            include_critical_curves=include_critical_curves,
            include_caustics=include_caustics,
            positions=positions,
            plot_in_kpc=plot_in_kpc,
            output_path=subplot_path,
            output_format="png",
        )

    ray_tracing_plotters.individual(
        tracer=tracer,
        grid=grid,
        mask=mask,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
        positions=positions,
        plot_profile_image=plot_image,
        plot_source_plane=plot_source_plane,
        plot_convergence=plot_convergence,
        plot_potential=plot_potential,
        plot_deflections=plot_deflections,
        plot_in_kpc=plot_in_kpc,
        output_path=output_path,
        output_format="png",
    )

    if not during_analysis:

        if plot_all_at_end_png:

            ray_tracing_plotters.individual(
                tracer=tracer,
                grid=grid,
                mask=mask,
                include_critical_curves=include_critical_curves,
                include_caustics=include_caustics,
                positions=positions,
                plot_profile_image=True,
                plot_source_plane=True,
                plot_convergence=True,
                plot_potential=True,
                plot_deflections=True,
                plot_in_kpc=plot_in_kpc,
                output_path=output_path,
                output_format="png",
            )

        if plot_all_at_end_fits:

            fits_path = af.path_util.make_and_return_path_from_path_and_folder_names(
                path=output_path, folder_names=["fits"]
            )

            ray_tracing_plotters.individual(
                tracer=tracer,
                grid=grid,
                mask=mask,
                include_critical_curves=include_critical_curves,
                include_caustics=include_caustics,
                positions=positions,
                plot_profile_image=True,
                plot_source_plane=True,
                plot_convergence=True,
                plot_potential=True,
                plot_deflections=True,
                output_path=fits_path,
                output_format="fits",
            )


def imaging_fit_of_phase(
    fit,
    during_analysis,
    plot_in_kpc,
    include_mask,
    include_positions,
    include_critical_curves,
    include_caustics,
    include_image_plane_pix,
    plot_all_at_end_png,
    plot_all_at_end_fits,
    plot_fit_as_subplot,
    plot_fit_of_planes_as_subplot,
    plot_inversion_as_subplot,
    plot_image,
    plot_noise_map,
    plot_signal_to_noise_map,
    plot_model_image,
    plot_residual_map,
    plot_normalized_residual_map,
    plot_chi_squared_map,
    plot_inversion_residual_map,
    plot_inversion_normalized_residual_map,
    plot_inversion_chi_squared_map,
    plot_inversion_regularization_weights,
    plot_subtracted_images_of_planes,
    plot_model_images_of_planes,
    plot_plane_images_of_planes,
    visualize_path,
    subplot_path,
):

    output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=visualize_path, folder_names=["fit"]
    )

    if plot_fit_as_subplot:

        fit_imaging_plotters.subplot(
            fit=fit,
            include_mask=include_mask,
            include_positions=include_positions,
            include_critical_curves=include_critical_curves,
            include_caustics=include_caustics,
            include_image_plane_pix=include_image_plane_pix,
            plot_in_kpc=plot_in_kpc,
            output_path=subplot_path,
            output_format="png",
        )

    if plot_fit_of_planes_as_subplot:

        fit_imaging_plotters.subplot_of_planes(
            fit=fit,
            include_mask=include_mask,
            include_positions=include_positions,
            include_critical_curves=include_critical_curves,
            include_caustics=include_caustics,
            include_image_plane_pix=include_image_plane_pix,
            plot_in_kpc=plot_in_kpc,
            output_path=subplot_path,
            output_format="png",
        )

    if plot_inversion_as_subplot and fit.inversion is not None:

        aa.plot.inversion.subplot(
            inversion=fit.inversion,
            mask=fit.mask,
            output_path=subplot_path,
            output_format="png",
        )

    fit_imaging_plotters.individuals(
        fit=fit,
        include_mask=include_mask,
        include_positions=include_positions,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
        include_image_plane_pix=include_image_plane_pix,
        plot_image=plot_image,
        plot_noise_map=plot_noise_map,
        plot_signal_to_noise_map=plot_signal_to_noise_map,
        plot_model_image=plot_model_image,
        plot_residual_map=plot_residual_map,
        plot_chi_squared_map=plot_chi_squared_map,
        plot_normalized_residual_map=plot_normalized_residual_map,
        plot_inversion_residual_map=plot_inversion_residual_map,
        plot_inversion_normalized_residual_map=plot_inversion_normalized_residual_map,
        plot_inversion_chi_squared_map=plot_inversion_chi_squared_map,
        plot_inversion_regularization_weight_map=plot_inversion_regularization_weights,
        plot_subtracted_images_of_planes=plot_subtracted_images_of_planes,
        plot_model_images_of_planes=plot_model_images_of_planes,
        plot_plane_images_of_planes=plot_plane_images_of_planes,
        plot_in_kpc=plot_in_kpc,
        output_path=output_path,
        output_format="png",
    )

    if not during_analysis:

        if plot_all_at_end_png:

            fit_imaging_plotters.individuals(
                fit=fit,
                include_mask=include_mask,
                include_positions=include_positions,
                include_critical_curves=include_critical_curves,
                include_caustics=include_caustics,
                include_image_plane_pix=include_image_plane_pix,
                plot_image=True,
                plot_noise_map=True,
                plot_signal_to_noise_map=True,
                plot_model_image=True,
                plot_residual_map=True,
                plot_normalized_residual_map=True,
                plot_chi_squared_map=True,
                plot_inversion_residual_map=True,
                plot_inversion_normalized_residual_map=True,
                plot_inversion_chi_squared_map=True,
                plot_inversion_regularization_weight_map=True,
                plot_subtracted_images_of_planes=True,
                plot_model_images_of_planes=True,
                plot_plane_images_of_planes=True,
                plot_in_kpc=plot_in_kpc,
                output_path=output_path,
                output_format="png",
            )

        if plot_all_at_end_fits:

            fits_path = af.path_util.make_and_return_path_from_path_and_folder_names(
                path=output_path, folder_names=["fits"]
            )

            fit_imaging_plotters.individuals(
                fit=fit,
                include_mask=include_mask,
                include_positions=include_positions,
                include_critical_curves=include_critical_curves,
                include_caustics=include_caustics,
                include_image_plane_pix=include_image_plane_pix,
                plot_image=True,
                plot_noise_map=True,
                plot_signal_to_noise_map=True,
                plot_model_image=True,
                plot_residual_map=True,
                plot_normalized_residual_map=True,
                plot_chi_squared_map=True,
                plot_inversion_residual_map=True,
                plot_inversion_normalized_residual_map=True,
                plot_inversion_chi_squared_map=True,
                plot_inversion_regularization_weight_map=True,
                plot_subtracted_images_of_planes=True,
                plot_model_images_of_planes=True,
                plot_plane_images_of_planes=True,
                output_path=fits_path,
                output_format="fits",
            )


def interferometer_fit_of_phase(
    fit,
    during_analysis,
    include_positions,
    include_mask,
    include_critical_curves,
    include_caustics,
    include_image_plane_pix,
    plot_in_kpc,
    plot_all_at_end_png,
    plot_all_at_end_fits,
    plot_fit_as_subplot,
    plot_inversion_as_subplot,
    plot_visibilities,
    plot_noise_map,
    plot_signal_to_noise_map,
    plot_model_visibilities,
    plot_residual_map,
    plot_normalized_residual_map,
    plot_chi_squared_map,
    plot_inversion_residual_map,
    plot_inversion_normalized_residual_map,
    plot_inversion_chi_squared_map,
    plot_inversion_regularization_weights,
    visualize_path,
    subplot_path,
):

    output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=visualize_path, folder_names=["fit"]
    )

    if plot_fit_as_subplot:

        fit_interferometer_plotters.subplot(
            fit=fit,
            plot_in_kpc=plot_in_kpc,
            output_path=subplot_path,
            output_format="png",
        )

        fit_interferometer_plotters.subplot_real_space(
            fit=fit,
            include_mask=include_mask,
            include_critical_curves=include_critical_curves,
            include_caustics=include_caustics,
            include_positions=include_positions,
            include_image_plane_pix=include_image_plane_pix,
            plot_in_kpc=plot_in_kpc,
            output_path=subplot_path,
            output_format="png",
        )

    # if plot_inversion_as_subplot and fit.inversion is not None:
    #
    #     aa.plot.inversion.subplot(
    #         inversion=fit.inversion,
    #         mask=fit.masked_interferometer.real_space_mask,
    #         positions=positions,
    #         output_path=subplot_path,
    #         output_format="png",
    #     )

    fit_interferometer_plotters.individuals(
        fit=fit,
        plot_visibilities=plot_visibilities,
        plot_noise_map=plot_noise_map,
        plot_signal_to_noise_map=plot_signal_to_noise_map,
        plot_model_visibilities=plot_model_visibilities,
        plot_residual_map=plot_residual_map,
        plot_chi_squared_map=plot_chi_squared_map,
        plot_normalized_residual_map=plot_normalized_residual_map,
        # plot_inversion_residual_map=plot_inversion_residual_map,
        # plot_inversion_normalized_residual_map=plot_inversion_normalized_residual_map,
        # plot_inversion_chi_squared_map=plot_inversion_chi_squared_map,
        plot_inversion_regularization_weight_map=plot_inversion_regularization_weights,
        plot_in_kpc=plot_in_kpc,
        output_path=output_path,
        output_format="png",
    )

    if not during_analysis:

        if plot_all_at_end_png:

            fit_interferometer_plotters.individuals(
                fit=fit,
                plot_visibilities=True,
                plot_noise_map=True,
                plot_signal_to_noise_map=True,
                plot_model_visibilities=True,
                plot_residual_map=True,
                plot_normalized_residual_map=True,
                plot_chi_squared_map=True,
                # plot_inversion_residual_map=True,
                # plot_inversion_normalized_residual_map=True,
                # plot_inversion_chi_squared_map=True,
                plot_inversion_regularization_weight_map=True,
                plot_in_kpc=plot_in_kpc,
                output_path=output_path,
                output_format="png",
            )

        if plot_all_at_end_fits:

            fits_path = af.path_util.make_and_return_path_from_path_and_folder_names(
                path=output_path, folder_names=["fits"]
            )

            fit_interferometer_plotters.individuals(
                fit=fit,
                plot_visibilities=True,
                plot_noise_map=True,
                plot_signal_to_noise_map=True,
                plot_model_visibilities=True,
                plot_residual_map=True,
                plot_normalized_residual_map=True,
                plot_chi_squared_map=True,
                # plot_inversion_residual_map=True,
                # plot_inversion_normalized_residual_map=True,
                # plot_inversion_chi_squared_map=True,
                plot_inversion_regularization_weight_map=True,
                output_path=fits_path,
                output_format="fits",
            )


def plot_hyper_images_for_phase(
    hyper_model_image,
    hyper_galaxy_image_path_dict,
    mask,
    kpc_per_arcsec,
    unit_label,
    plot_hyper_model_image,
    plot_hyper_galaxy_images,
    visualize_path,
):

    output_path = af.path_util.make_and_return_path_from_path_and_folder_names(
        path=visualize_path, folder_names=["hyper_galaxies"]
    )

    if plot_hyper_model_image:

        hyper_plotters.hyper_model_image(
            hyper_model_image=hyper_model_image,
            mask=mask,
            kpc_per_arcsec=kpc_per_arcsec,
            unit_label=unit_label,
            output_path=output_path,
            output_format="png",
        )

    if plot_hyper_galaxy_images:

        hyper_plotters.subplot_of_hyper_galaxy_images(
            hyper_galaxy_image_path_dict=hyper_galaxy_image_path_dict,
            mask=mask,
            kpc_per_arcsec=kpc_per_arcsec,
            unit_label=unit_label,
            output_path=output_path,
            output_format="png",
        )



import numpy as np

import autofit as af
import autoarray as aa
from autoastro.galaxy import galaxy as g
from autolens.pipeline.phase import dataset


class Result(dataset.Result):
    @property
    def most_likely_fit(self):

        hyper_background_noise = self.analysis.hyper_background_noise_for_instance(
            instance=self.instance
        )

        return self.analysis.masked_interferometer_fit_for_tracer(
            tracer=self.most_likely_tracer,
            hyper_background_noise=hyper_background_noise,
        )

    @property
    def unmasked_model_visibilities(self):
        return self.most_likely_fit.unmasked_blurred_profile_image

    @property
    def unmasked_model_visibilities_of_planes(self):
        return self.most_likely_fit.unmasked_blurred_profile_image_of_planes

    @property
    def unmasked_model_visibilities_of_planes_and_galaxies(self):
        fit = self.most_likely_fit
        return fit.unmasked_blurred_profile_image_of_planes_and_galaxies

    def visibilities_for_galaxy(self, galaxy: g.Galaxy) -> np.ndarray:
        """
        Parameters
        ----------
        galaxy
            A galaxy used in this phase

        Returns
        -------
        ndarray or None
            A numpy arrays giving the model visibilities of that galaxy
        """
        return self.most_likely_fit.galaxy_model_visibilities_dict[galaxy]

    @property
    def visibilities_galaxy_dict(self) -> {str: g.Galaxy}:
        """
        A dictionary associating galaxy names with model visibilities of those galaxies
        """
        return {
            galaxy_path: self.visibilities_for_galaxy(galaxy)
            for galaxy_path, galaxy in self.path_galaxy_tuples
        }

    @property
    def hyper_galaxy_visibilities_path_dict(self):
        """
        A dictionary associating 1D hyper_galaxies galaxy visibilities with their names.
        """

        hyper_minimum_percent = af.conf.instance.general.get(
            "hyper", "hyper_minimum_percent", float
        )

        hyper_galaxy_visibilities_path_dict = {}

        for path, galaxy in self.path_galaxy_tuples:

            galaxy_visibilities = self.visibilities_galaxy_dict[path]

            if not np.all(galaxy_visibilities == 0):
                minimum_galaxy_value = hyper_minimum_percent * max(galaxy_visibilities)
                galaxy_visibilities[
                    galaxy_visibilities < minimum_galaxy_value
                ] = minimum_galaxy_value

            hyper_galaxy_visibilities_path_dict[path] = galaxy_visibilities

        return hyper_galaxy_visibilities_path_dict

    @property
    def hyper_model_visibilities(self):

        hyper_model_visibilities = aa.masked.array.zeros(mask=self.mask.mask_sub_1)

        for path, galaxy in self.path_galaxy_tuples:
            hyper_model_visibilities += self.hyper_galaxy_visibilities_path_dict[path]

        return hyper_model_visibilities

    def image_for_galaxy(self, galaxy: g.Galaxy) -> np.ndarray:
        """
        Parameters
        ----------
        galaxy
            A galaxy used in this phase

        Returns
        -------
        ndarray or None
            A numpy arrays giving the model image of that galaxy
        """
        return self.most_likely_fit.galaxy_model_image_dict[galaxy]

    @property
    def image_galaxy_dict(self) -> {str: g.Galaxy}:
        """
        A dictionary associating galaxy names with model images of those galaxies
        """
        return {
            galaxy_path: self.image_for_galaxy(galaxy)
            for galaxy_path, galaxy in self.path_galaxy_tuples
        }

    @property
    def hyper_galaxy_image_path_dict(self):
        """
        A dictionary associating 1D hyper_galaxies galaxy images with their names.
        """

        hyper_minimum_percent = af.conf.instance.general.get(
            "hyper", "hyper_minimum_percent", float
        )

        hyper_galaxy_image_path_dict = {}

        for path, galaxy in self.path_galaxy_tuples:

            galaxy_image = self.image_galaxy_dict[path]

            if not np.all(galaxy_image == 0):
                minimum_galaxy_value = hyper_minimum_percent * max(galaxy_image)
                galaxy_image[galaxy_image < minimum_galaxy_value] = minimum_galaxy_value

            hyper_galaxy_image_path_dict[path] = galaxy_image

        return hyper_galaxy_image_path_dict

    @property
    def hyper_model_image(self):

        hyper_model_image = aa.masked.array.zeros(mask=self.mask.mask_sub_1)

        for path, galaxy in self.path_galaxy_tuples:
            hyper_model_image += self.hyper_galaxy_image_path_dict[path]

        return hyper_model_image



class PipelineTuple(object):
    def __init__(self, module):
        self.make = module.make
        self.__doc__ = module.__doc__

    @property
    def short_doc(self):
        return self.__doc__.split("\n")[1]

    @property
    def doc(self):
        return self.__doc__.replace("  ", "").replace("\n", " ")


pipeline_dict = {}


def add(module):
    """
    Parameters
    ----------
    module: {analysis_path, make}
    """
    pipeline_dict[module.pipeline_name] = PipelineTuple(module)


#  Add pipeline modules here.
# add(initialize)


class TestPipeline(object):
    # noinspection PyMethodMayBeStatic
    def run(self, dataset):
        print(dataset)



from astropy import cosmology as cosmo

import autofit as af
import autoarray as aa
from autolens.pipeline.phase import abstract
from autolens.pipeline.phase import extensions
from autolens.pipeline.phase.dataset.result import Result


def default_mask_function(shape_2d, pixel_scales):
    return aa.mask.circular(
        shape_2d=shape_2d, pixel_scales=pixel_scales, sub_size=1, radius=3.0
    )


def isinstance_or_prior(obj, cls):
    if isinstance(obj, cls):
        return True
    if isinstance(obj, af.PriorModel) and obj.cls == cls:
        return True
    return False


class PhaseDataset(abstract.AbstractPhase):
    galaxies = af.PhaseProperty("galaxies")

    Result = Result

    @af.convert_paths
    def __init__(
        self,
        paths,
        galaxies=None,
        optimizer_class=af.MultiNest,
        cosmology=cosmo.Planck15,
    ):
        """

        A phase in an lens pipeline. Uses the set non_linear optimizer to try to fit models and hyper_galaxies
        passed to it.

        Parameters
        ----------
        optimizer_class: class
            The class of a non_linear optimizer
        """

        super(PhaseDataset, self).__init__(paths, optimizer_class=optimizer_class)
        self.galaxies = galaxies or []
        self.cosmology = cosmology

        self.is_hyper_phase = False

    def run(self, dataset, results=None, mask=None, positions=None):
        """
        Run this phase.

        Parameters
        ----------
        positions
        mask: Mask
            The default masks passed in by the pipeline
        results: autofit.tools.pipeline.ResultsCollection
            An object describing the results of the last phase or None if no phase has been executed
        dataset: scaled_array.ScaledSquarePixelArray
            An masked_imaging that has been masked

        Returns
        -------
        result: AbstractPhase.Result
            A result object comprising the best fit model and other hyper_galaxies.
        """
        self.model = self.model.populate(results)

        analysis = self.make_analysis(
            dataset=dataset, results=results, mask=mask, positions=positions
        )

        self.customize_priors(results)
        self.assert_and_save_pickle()

        result = self.run_analysis(analysis)

        return self.make_result(result=result, analysis=analysis)

    def make_analysis(self, dataset, results=None, mask=None, positions=None):
        """
        Create an lens object. Also calls the prior passing and masked_imaging modifying functions to allow child
        classes to change the behaviour of the phase.

        Parameters
        ----------
        positions
        mask: Mask
            The default masks passed in by the pipeline
        dataset: im.Imaging
            An masked_imaging that has been masked
        results: autofit.tools.pipeline.ResultsCollection
            The result from the previous phase

        Returns
        -------
        lens : Analysis
            An lens object that the non-linear optimizer calls to determine the fit of a set of values
        """
        raise NotImplementedError()

    def extend_with_inversion_phase(self):
        return extensions.InversionPhase(phase=self)



import autofit as af
import autolens as al
from test_autolens.integration.tests.imaging import runner

test_type = "regression"
test_name = "new_api_pixelization"
data_type = "lens_mass__source_smooth"
data_resolution = "lsst"


def make_pipeline(
    name,
    phase_folders,
    pipeline_pixelization=al.pix.VoronoiBrightnessImage,
    pipeline_regularization=al.reg.AdaptiveBrightness,
    optimizer_class=af.MultiNest,
):
    phase1 = al.PhaseImaging(
        phase_name="phase_1__lens_sersic",
        phase_folders=phase_folders,
        galaxies=dict(
            source=al.GalaxyModel(redshift=1.0, light=al.lp.SphericalExponential),
            lens=al.GalaxyModel(redshift=0.5, light=al.mp.SphericalIsothermal()),
        ),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 20
    phase1.optimizer.sampling_efficiency = 0.8

    phase1 = phase1.extend_with_multiple_hyper_phases(
        hyper_galaxy=True, include_background_sky=True, include_background_noise=True
    )

    phase6 = al.PhaseImaging(
        phase_name="phase_6_initialize_inversion",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                mass=al.mp.SphericalIsothermal(),
                hyper_galaxy=phase1.result.hyper_combined.instance.galaxies.lens.hyper_galaxy,
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=pipeline_pixelization,
                regularization=pipeline_regularization,
            ),
        ),
        hyper_image_sky=phase1.result.hyper_combined.instance.hyper_image_sky,
        hyper_background_noise=phase1.result.hyper_combined.instance.hyper_background_noise,
        optimizer_class=optimizer_class,
    )

    phase6.optimizer.const_efficiency_mode = True
    phase6.optimizer.n_live_points = 20
    phase6.optimizer.sampling_efficiency = 0.8

    phase6 = phase6.extend_with_multiple_hyper_phases(
        hyper_galaxy=True,
        include_background_sky=True,
        include_background_noise=True,
        inversion=True,
    )

    phase7 = al.PhaseImaging(
        phase_name="phase_7__lens_sersic_sie__source_inversion",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                mass=al.mp.SphericalExponential(),
                hyper_galaxy=phase6.result.hyper_combined.instance.galaxies.lens.hyper_galaxy,
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                pixelization=phase6.result.instance.galaxies.source.pixelization,
                regularization=phase6.result.instance.galaxies.source.regularization,
                hyper_galaxy=phase6.result.hyper_combined.instance.galaxies.source.hyper_galaxy,
            ),
        ),
        hyper_image_sky=phase1.result.hyper_combined.instance.hyper_image_sky,
        hyper_background_noise=phase1.result.hyper_combined.instance.hyper_background_noise,
        optimizer_class=optimizer_class,
    )

    phase7.optimizer.const_efficiency_mode = True
    phase7.optimizer.n_live_points = 75
    phase7.optimizer.sampling_efficiency = 0.2

    phase7 = phase7.extend_with_multiple_hyper_phases(
        hyper_galaxy=True,
        include_background_sky=True,
        include_background_noise=True,
        inversion=True,
    )

    return al.PipelineDataset(name, phase1, phase6, phase7, hyper_mode=False)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



import autofit as af
import matplotlib

backend = af.conf.get_matplotlib_backend()
matplotlib.use(backend)
from matplotlib import pyplot as plt

import autoarray as aa
from autoarray.util import plotter_util
from autoastro.plotters import lens_plotter_util
from autolens.plotters import plane_plotters


def subplot(
    tracer,
    grid,
    mask=None,
    include_critical_curves=False,
    include_caustics=False,
    positions=None,
    plot_in_kpc=False,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1.0,
    output_path=None,
    output_filename="tracer",
    output_format="show",
):
    """Plot the observed _tracer of an analysis, using the *Imaging* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    tracer : autolens.imaging.tracer.Imaging
        Class containing the _tracer,  noise_map-mappers and PSF that are to be plotted.
        The font size of the figure ylabel.
    output_path : str
        The path where the _tracer is output if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the _tracer is output. File formats (e.g. png, fits) output the _tracer to harddisk. 'show' displays the _tracer \
        in the python interpreter window.
    """

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=6
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)
    plt.subplot(rows, columns, 1)

    profile_image(
        tracer=tracer,
        grid=grid,
        mask=mask,
        include_critical_curves=include_critical_curves,
        positions=positions,
        as_subplot=True,
        plot_in_kpc=plot_in_kpc,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    if tracer.has_mass_profile:

        plt.subplot(rows, columns, 2)

        convergence(
            tracer=tracer,
            grid=grid,
            mask=mask,
            as_subplot=True,
            plot_in_kpc=plot_in_kpc,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            output_path=output_path,
            output_filename="",
            output_format=output_format,
        )

        plt.subplot(rows, columns, 3)

        potential(
            tracer=tracer,
            grid=grid,
            mask=mask,
            as_subplot=True,
            plot_in_kpc=plot_in_kpc,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            output_path=output_path,
            output_filename="",
            output_format=output_format,
        )

    plt.subplot(rows, columns, 4)

    source_plane_grid = tracer.traced_grids_of_planes_from_grid(grid=grid)[-1]

    caustics = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=tracer, include_critical_curves=False, include_caustics=include_caustics
    )

    plane_plotters.plane_image(
        plane=tracer.source_plane,
        grid=source_plane_grid,
        lines=caustics,
        as_subplot=True,
        positions=None,
        include_grid=False,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_filename="",
        output_format=output_format,
    )

    if tracer.has_mass_profile:

        plt.subplot(rows, columns, 5)

        deflections_y(
            tracer=tracer,
            grid=grid,
            mask=mask,
            as_subplot=True,
            plot_in_kpc=plot_in_kpc,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            output_path=output_path,
            output_filename="",
            output_format=output_format,
        )

        plt.subplot(rows, columns, 6)

        deflections_x(
            tracer=tracer,
            grid=grid,
            mask=mask,
            as_subplot=True,
            plot_in_kpc=plot_in_kpc,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            output_path=output_path,
            output_filename="",
            output_format=output_format,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    plt.close()


def individual(
    tracer,
    grid,
    mask=None,
    include_critical_curves=False,
    include_caustics=False,
    positions=None,
    plot_profile_image=False,
    plot_source_plane=False,
    plot_convergence=False,
    plot_potential=False,
    plot_deflections=False,
    plot_in_kpc=False,
    output_path=None,
    output_format="show",
):
    """Plot the observed _tracer of an analysis, using the *Imaging* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    tracer : autolens.imaging.tracer.Imaging
        Class containing the _tracer, noise_map-mappers and PSF that are to be plotted.
        The font size of the figure ylabel.
    output_path : str
        The path where the _tracer is output if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the _tracer is output. File formats (e.g. png, fits) output the _tracer to harddisk. 'show' displays the _tracer \
        in the python interpreter window.
    """

    if plot_profile_image:

        profile_image(
            tracer=tracer,
            grid=grid,
            mask=mask,
            include_critical_curves=include_critical_curves,
            positions=positions,
            plot_in_kpc=plot_in_kpc,
            output_path=output_path,
            output_format=output_format,
        )

    if plot_convergence:

        convergence(
            tracer=tracer,
            grid=grid,
            mask=mask,
            plot_in_kpc=plot_in_kpc,
            output_path=output_path,
            output_format=output_format,
        )

    if plot_potential:

        potential(
            tracer=tracer,
            grid=grid,
            mask=mask,
            plot_in_kpc=plot_in_kpc,
            output_path=output_path,
            output_format=output_format,
        )

    if plot_source_plane:

        source_plane_grid = tracer.traced_grids_of_planes_from_grid(grid=grid)[-1]

        caustics = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
            obj=tracer, include_critical_curves=False, include_caustics=include_caustics
        )

        plane_plotters.plane_image(
            plane=tracer.source_plane,
            grid=source_plane_grid,
            lines=caustics,
            positions=None,
            include_grid=False,
            plot_in_kpc=plot_in_kpc,
            output_path=output_path,
            output_filename="tracer_source_plane",
            output_format=output_format,
        )

    if plot_deflections:

        deflections_y(
            tracer=tracer,
            grid=grid,
            mask=mask,
            plot_in_kpc=plot_in_kpc,
            output_path=output_path,
            output_format=output_format,
        )

    if plot_deflections:

        deflections_x(
            tracer=tracer,
            grid=grid,
            mask=mask,
            plot_in_kpc=plot_in_kpc,
            output_path=output_path,
            output_format=output_format,
        )


def profile_image(
    tracer,
    grid,
    mask=None,
    include_critical_curves=False,
    include_caustics=False,
    positions=None,
    as_subplot=False,
    plot_in_kpc=False,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Tracer Profile Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="tracer_profile_image",
):

    profile_image = tracer.profile_image_from_grid(grid=grid)

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=tracer,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=tracer.image_plane, plot_in_kpc=plot_in_kpc
    )

    aa.plot.array(
        array=profile_image,
        mask=mask,
        lines=lines,
        points=positions,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def convergence(
    tracer,
    grid,
    mask=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    plot_in_kpc=False,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Tracer Convergence",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="tracer_convergence",
):

    convergence = tracer.convergence_from_grid(grid=grid)

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=tracer,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=tracer.image_plane, plot_in_kpc=plot_in_kpc
    )

    aa.plot.array(
        array=convergence,
        mask=mask,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def potential(
    tracer,
    grid,
    mask=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    plot_in_kpc=False,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Tracer Potential",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="tracer_potential",
):

    potential = tracer.potential_from_grid(grid=grid)

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=tracer,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=tracer.image_plane, plot_in_kpc=plot_in_kpc
    )

    aa.plot.array(
        array=potential,
        mask=mask,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def deflections_y(
    tracer,
    grid,
    mask=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    plot_in_kpc=False,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Tracer Deflections (y)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="tracer_deflections_y",
):

    deflections = tracer.deflections_from_grid(grid=grid)
    deflections_y = grid.mapping.array_from_sub_array_1d(sub_array_1d=deflections[:, 0])

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=tracer,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=tracer.image_plane, plot_in_kpc=plot_in_kpc
    )

    aa.plot.array(
        array=deflections_y,
        mask=mask,
        lines=lines,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def deflections_x(
    tracer,
    grid,
    mask=None,
    include_critical_curves=False,
    include_caustics=False,
    as_subplot=False,
    plot_in_kpc=False,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Tracer Deflections (x)",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    output_path=None,
    output_format="show",
    output_filename="tracer_deflections_x",
):

    deflections = tracer.deflections_from_grid(grid=grid)
    deflections_x = grid.mapping.array_from_sub_array_1d(sub_array_1d=deflections[:, 1])

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=tracer,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=tracer.image_plane, plot_in_kpc=plot_in_kpc
    )

    aa.plot.array(
        array=deflections_x,
        lines=lines,
        mask=mask,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )



import autolens as al
import os

import pytest


@pytest.fixture(name="ray_tracing_plotter_path")
def make_ray_tracing_plotter_setup():
    return "{}/../../test_files/plotting/ray_tracing/".format(
        os.path.dirname(os.path.realpath(__file__))
    )


def test__all_individual_plotters(
    tracer_x2_plane_7x7, sub_grid_7x7, mask_7x7, ray_tracing_plotter_path, plot_patch
):
    al.plot.tracer.profile_image(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=ray_tracing_plotter_path,
        output_format="png",
    )

    assert ray_tracing_plotter_path + "tracer_profile_image.png" in plot_patch.paths

    al.plot.tracer.convergence(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=ray_tracing_plotter_path,
        output_format="png",
    )

    assert ray_tracing_plotter_path + "tracer_convergence.png" in plot_patch.paths

    al.plot.tracer.potential(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=ray_tracing_plotter_path,
        output_format="png",
    )

    assert ray_tracing_plotter_path + "tracer_potential.png" in plot_patch.paths

    al.plot.tracer.deflections_y(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=ray_tracing_plotter_path,
        output_format="png",
    )

    assert ray_tracing_plotter_path + "tracer_deflections_y.png" in plot_patch.paths

    al.plot.tracer.deflections_x(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=ray_tracing_plotter_path,
        output_format="png",
    )

    assert ray_tracing_plotter_path + "tracer_deflections_x.png" in plot_patch.paths


def test__tracer_sub_plot_output(
    tracer_x2_plane_7x7, sub_grid_7x7, mask_7x7, ray_tracing_plotter_path, plot_patch
):
    al.plot.tracer.subplot(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        output_path=ray_tracing_plotter_path,
        output_format="png",
    )

    assert ray_tracing_plotter_path + "tracer.png" in plot_patch.paths


def test__tracer_individuals__dependent_on_input(
    tracer_x2_plane_7x7, sub_grid_7x7, mask_7x7, ray_tracing_plotter_path, plot_patch
):
    al.plot.tracer.individual(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        plot_profile_image=True,
        plot_source_plane=True,
        plot_potential=True,
        output_path=ray_tracing_plotter_path,
        output_format="png",
    )

    assert ray_tracing_plotter_path + "tracer_profile_image.png" in plot_patch.paths

    assert ray_tracing_plotter_path + "tracer_source_plane.png" in plot_patch.paths

    assert ray_tracing_plotter_path + "tracer_convergence.png" not in plot_patch.paths

    assert ray_tracing_plotter_path + "tracer_potential.png" in plot_patch.paths

    assert ray_tracing_plotter_path + "tracer_deflections_y.png" not in plot_patch.paths

    assert ray_tracing_plotter_path + "tracer_deflections_x.png" not in plot_patch.paths



import os

import autofit as af
from test_autolens.integration import integration_util
from test_autolens.simulate.imaging import simulate_util
from autofit.optimize.non_linear.mock_nlo import MockNLO


def run(
    module,
    test_name=None,
    optimizer_class=af.MultiNest,
    config_folder="config",
    positions=None,
):
    test_name = test_name or module.test_name
    test_path = "{}/../../".format(os.path.dirname(os.path.realpath(__file__)))
    output_path = test_path + "output/imaging/"
    config_path = test_path + config_folder
    af.conf.instance = af.conf.Config(config_path=config_path, output_path=output_path)
    integration_util.reset_paths(test_name=test_name, output_path=output_path)

    imaging = simulate_util.load_test_imaging(
        data_type=module.data_type, data_resolution=module.data_resolution
    )

    module.make_pipeline(
        name=test_name,
        phase_folders=[module.test_type, test_name],
        optimizer_class=optimizer_class,
    ).run(dataset=imaging, positions=positions)


def run_a_mock(module):
    # noinspection PyTypeChecker
    run(
        module,
        test_name=f"{module.test_name}_mock",
        optimizer_class=MockNLO,
        config_folder="config_mock",
    )


def run_with_multi_nest(module):
    # noinspection PyTypeChecker
    run(
        module,
        test_name=f"{module.test_name}_nest",
        optimizer_class=af.MultiNest,
        config_folder="config_mock",
    )



from autoarray.plotters.fit_imaging_plotters import *
from autolens.plotters.fit_imaging_plotters.fit_imaging_plotters import (
    subplot,
    subplot_for_plane,
    subplot_of_planes,
    subtracted_image_of_plane,
    individuals,
    model_image_of_plane,
    contribution_maps,
)



import autolens as al
import pytest
import os


@pytest.fixture(name="plane_plotter_path")
def make_plane_plotter_setup():
    return "{}/../../test_files/plotting/plane/".format(
        os.path.dirname(os.path.realpath(__file__))
    )


def test__all_individual_plotters__output_file_with_default_name(
    plane_7x7, sub_grid_7x7, mask_7x7, positions_7x7, plane_plotter_path, plot_patch
):

    al.plot.plane.profile_image(
        plane=plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        positions=positions_7x7,
        plot_in_kpc=True,
        include_critical_curves=True,
        include_caustics=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=plane_plotter_path,
        output_format="png",
    )

    assert plane_plotter_path + "plane_profile_image.png" in plot_patch.paths

    al.plot.plane.plane_image(
        plane=plane_7x7,
        grid=sub_grid_7x7,
        positions=positions_7x7,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=plane_plotter_path,
        output_format="png",
    )

    assert plane_plotter_path + "plane_plane_image.png" in plot_patch.paths

    al.plot.plane.convergence(
        plane=plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        include_critical_curves=True,
        include_caustics=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=plane_plotter_path,
        output_format="png",
    )

    assert plane_plotter_path + "plane_convergence.png" in plot_patch.paths

    al.plot.plane.potential(
        plane=plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        include_critical_curves=True,
        include_caustics=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=plane_plotter_path,
        output_format="png",
    )

    assert plane_plotter_path + "plane_potential.png" in plot_patch.paths

    al.plot.plane.deflections_y(
        plane=plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        include_critical_curves=True,
        include_caustics=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=plane_plotter_path,
        output_format="png",
    )

    assert plane_plotter_path + "plane_deflections_y.png" in plot_patch.paths

    al.plot.plane.deflections_x(
        plane=plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        include_critical_curves=True,
        include_caustics=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=plane_plotter_path,
        output_format="png",
    )

    assert plane_plotter_path + "plane_deflections_x.png" in plot_patch.paths

    al.plot.plane.magnification(
        plane=plane_7x7,
        grid=sub_grid_7x7,
        mask=mask_7x7,
        include_critical_curves=True,
        include_caustics=True,
        cb_tick_values=[1.0],
        cb_tick_labels=["1.0"],
        output_path=plane_plotter_path,
        output_format="png",
    )

    assert plane_plotter_path + "plane_magnification.png" in plot_patch.paths

    al.plot.plane.plane_grid(
        plane=plane_7x7,
        grid=sub_grid_7x7,
        output_path=plane_plotter_path,
        output_format="png",
    )

    assert plane_plotter_path + "plane_grid.png" in plot_patch.paths



import autofit as af
from autolens.lens import ray_tracing


class Analysis(af.Analysis):
    def __init__(self, cosmology):

        self.cosmology = cosmology

    def hyper_image_sky_for_instance(self, instance):

        if hasattr(instance, "hyper_image_sky"):
            return instance.hyper_image_sky
        else:
            return None

    def hyper_background_noise_for_instance(self, instance):

        if hasattr(instance, "hyper_background_noise"):
            return instance.hyper_background_noise
        else:
            return None

    def tracer_for_instance(self, instance):
        return ray_tracing.Tracer.from_galaxies(
            galaxies=instance.galaxies, cosmology=self.cosmology
        )



import autofit as af
import autolens as al
from test_autolens.integration.tests.imaging import runner

test_type = "lens__source"
test_name = "lens_light_mass__source__hyper_bg"
data_type = "lens_light__source_smooth"
data_resolution = "lsst"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):

    phase1 = al.PhaseImaging(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=al.lp.SphericalDevVaucouleurs,
                mass=al.mp.EllipticalIsothermal,
            ),
            source=al.GalaxyModel(redshift=1.0, light=al.lp.EllipticalSersic),
        ),
        optimizer_class=optimizer_class,
    )

    phase1.optimizer.const_efficiency_mode = True
    phase1.optimizer.n_live_points = 60
    phase1.optimizer.sampling_efficiency = 0.8

    phase1 = phase1.extend_with_multiple_hyper_phases(
        hyper_galaxy=True, include_background_sky=True, include_background_noise=True
    )

    phase2 = al.PhaseImaging(
        phase_name="phase_2",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(
                redshift=0.5,
                light=phase1.result.model.galaxies.lens.light,
                mass=phase1.result.model.galaxies.lens.mass,
                hyper_galaxy=phase1.result.hyper_combined.instance.galaxies.lens.hyper_galaxy,
            ),
            source=al.GalaxyModel(
                redshift=1.0,
                light=phase1.result.model.galaxies.source.light,
                hyper_galaxy=phase1.result.hyper_combined.instance.galaxies.source.hyper_galaxy,
            ),
        ),
        hyper_image_sky=phase1.result.hyper_combined.instance.hyper_image_sky,
        hyper_background_noise=phase1.result.hyper_combined.instance.hyper_background_noise,
        optimizer_class=optimizer_class,
    )

    phase2.optimizer.const_efficiency_mode = True
    phase2.optimizer.n_live_points = 40
    phase2.optimizer.sampling_efficiency = 0.8

    phase2 = phase2.extend_with_multiple_hyper_phases(
        hyper_galaxy=True, include_background_sky=True, include_background_noise=True
    )

    return al.PipelineDataset(name, phase1, phase2)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



import autofit as af
import matplotlib

backend = af.conf.get_matplotlib_backend()
matplotlib.use(backend)
from matplotlib import pyplot as plt

import autoarray as aa
from autoarray.util import plotter_util
from autoastro.plotters import lens_plotter_util
from autolens.plotters import plane_plotters, ray_tracing_plotters


def subplot(
    fit,
    plot_in_kpc=False,
    figsize=None,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    grid_pointsize=1,
    output_path=None,
    output_filename="fit",
    output_format="show",
):

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=6
    )

    if figsize is None:
        figsize = figsize_tool

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=fit.tracer.image_plane, plot_in_kpc=plot_in_kpc
    )

    plt.figure(figsize=figsize)

    plt.subplot(rows, columns, 1)

    aa.plot.fit_interferometer.subplot(
        fit=fit,
        unit_label=unit_label,
        unit_conversion_factor=unit_conversion_factor,
        figsize=figsize,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )


def subplot_real_space(
    fit,
    include_mask=True,
    include_critical_curves=False,
    include_caustics=False,
    include_positions=True,
    include_image_plane_pix=False,
    include_mass_profile_centres=True,
    plot_in_kpc=False,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10,
    grid_pointsize=1,
    output_path=None,
    output_filename="fit_real_space",
    output_format="show",
):

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=2
    )

    if figsize is None:
        figsize = figsize_tool

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=fit.tracer.image_plane, plot_in_kpc=plot_in_kpc
    )

    image_plane_pix_grid = lens_plotter_util.get_image_plane_pix_grid_from_fit(
        include_image_plane_pix=include_image_plane_pix, fit=fit
    )

    real_space_mask = plotter_util.get_real_space_mask_from_fit(fit=fit, include_mask=include_mask)

    positions = lens_plotter_util.get_positions_from_fit(fit=fit, include_positions=include_positions)

    plt.figure(figsize=figsize)

    lines = lens_plotter_util.get_critical_curves_and_caustics_from_lensing_object(
        obj=fit.tracer,
        include_critical_curves=include_critical_curves,
        include_caustics=include_caustics,
    )

    plt.subplot(rows, columns, 1)

    if not fit.inversion is not None:

        ray_tracing_plotters.profile_image(
            tracer=fit.tracer,
            grid=fit.masked_interferometer.grid,
            mask=real_space_mask,
            include_critical_curves=include_critical_curves,
            positions=positions,
            as_subplot=True,
            plot_in_kpc=plot_in_kpc,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            position_pointsize=position_pointsize,
            mask_pointsize=mask_pointsize,
            output_path=output_path,
            output_filename="",
            output_format=output_format,
        )

        plt.subplot(rows, columns, 2)

        plane_plotters.plane_image(
            plane=fit.tracer.source_plane,
            grid=fit.masked_interferometer.grid,
            as_subplot=True,
            lines=[lines[1]],
            plot_in_kpc=plot_in_kpc,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            grid_pointsize=grid_pointsize,
            position_pointsize=position_pointsize,
            output_path=output_path,
            output_filename="",
            output_format=output_format,
        )

    elif fit.inversion is not None:

        aa.plot.inversion.reconstructed_image(
            inversion=fit.inversion,
            mask=real_space_mask,
            lines=lines[0],
            positions=positions,
            grid=image_plane_pix_grid,
            as_subplot=True,
            unit_label=unit_label,
            unit_conversion_factor=unit_conversion_factor,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            output_path=output_path,
            output_format=output_format,
            output_filename=output_filename,
        )

        ratio = float(
            (
                fit.inversion.mapper.grid.scaled_maxima[1]
                - fit.inversion.mapper.grid.scaled_minima[1]
            )
            / (
                fit.inversion.mapper.grid.scaled_maxima[0]
                - fit.inversion.mapper.grid.scaled_minima[0]
            )
        )

        if aspect is "square":
            aspect_inv = ratio
        elif aspect is "auto":
            aspect_inv = 1.0 / ratio
        elif aspect is "equal":
            aspect_inv = 1.0

        plt.subplot(rows, columns, 2, aspect=float(aspect_inv))

        aa.plot.inversion.reconstruction(
            inversion=fit.inversion,
            lines=lines[0],
            include_grid=False,
            include_centres=False,
            as_subplot=True,
            unit_label=unit_label,
            unit_conversion_factor=unit_conversion_factor,
            figsize=figsize,
            aspect=None,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            output_path=output_path,
            output_filename=None,
            output_format=output_format,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    plt.close()


def individuals(
    fit,
    plot_in_kpc=False,
    plot_visibilities=False,
    plot_noise_map=False,
    plot_signal_to_noise_map=False,
    plot_model_visibilities=False,
    plot_residual_map=False,
    plot_normalized_residual_map=False,
    plot_chi_squared_map=False,
    plot_inversion_residual_map=False,
    plot_inversion_normalized_residual_map=False,
    plot_inversion_chi_squared_map=False,
    plot_inversion_regularization_weight_map=False,
    output_path=None,
    output_format="show",
):
    """Plot the model datas_ of an analysis, using the *Fitter* class object.

    The visualization and output type can be fully customized.

    Parameters
    -----------
    fit : autolens.lens.fitting.Fitter
        Class containing fit between the model datas_ and observed lens datas_ (including residual_map, chi_squared_map etc.)
    output_path : str
        The path where the datas_ is output if the output_type is a file format (e.g. png, fits)
    output_format : str
        How the datas_ is output. File formats (e.g. png, fits) output the datas_ to harddisk. 'show' displays the datas_ \
        in the python interpreter window.
    """

    unit_label, unit_conversion_factor = lens_plotter_util.get_unit_label_and_unit_conversion_factor(
        obj=fit.tracer.image_plane, plot_in_kpc=plot_in_kpc
    )

    aa.plot.fit_interferometer.individuals(
        fit=fit,
        plot_visibilities=plot_visibilities,
        plot_noise_map=plot_noise_map,
        plot_signal_to_noise_map=plot_signal_to_noise_map,
        plot_model_visibilities=plot_model_visibilities,
        plot_residual_map=plot_residual_map,
        plot_normalized_residual_map=plot_normalized_residual_map,
        plot_chi_squared_map=plot_chi_squared_map,
        plot_inversion_residual_map=plot_inversion_residual_map,
        plot_inversion_normalized_residual_map=plot_inversion_normalized_residual_map,
        plot_inversion_chi_squared_map=plot_inversion_chi_squared_map,
        plot_inversion_regularization_weight_map=plot_inversion_regularization_weight_map,
        unit_conversion_factor=unit_conversion_factor,
        unit_label=unit_label,
        output_path=output_path,
        output_format=output_format,
    )



import autolens as al
import os

import pytest


@pytest.fixture(name="phase_plotter_path")
def make_phase_plotter_setup():
    return "{}/../../test_files/plotting/phase/".format(
        os.path.dirname(os.path.realpath(__file__))
    )


def test__plot_imaging_for_phase(
    imaging_7x7, mask_7x7, general_config, phase_plotter_path, plot_patch
):
    al.plot.phase.imaging_of_phase(
        imaging=imaging_7x7,
        mask=mask_7x7,
        positions=None,
        kpc_per_arcsec=2.0,
        unit_label="kpc",
        plot_as_subplot=True,
        plot_image=True,
        plot_noise_map=False,
        plot_psf=True,
        plot_signal_to_noise_map=False,
        plot_absolute_signal_to_noise_map=False,
        plot_potential_chi_squared_map=True,
        visualize_path=phase_plotter_path,
        subplot_path=phase_plotter_path,
    )

    assert phase_plotter_path + "imaging.png" in plot_patch.paths
    assert phase_plotter_path + "imaging/imaging_image.png" in plot_patch.paths
    assert phase_plotter_path + "imaging/imaging_noise_map.png" not in plot_patch.paths
    assert phase_plotter_path + "imaging/imaging_psf.png" in plot_patch.paths
    assert (
        phase_plotter_path + "imaging/imaging_signal_to_noise_map.png"
        not in plot_patch.paths
    )
    assert (
        phase_plotter_path + "imaging/imaging_absolute_signal_to_noise_map.png"
        not in plot_patch.paths
    )
    assert (
        phase_plotter_path + "imaging/imaging_potential_chi_squared_map.png"
        in plot_patch.paths
    )


def test__plot_interferometer_for_phase(
    interferometer_7, mask_7x7, general_config, phase_plotter_path, plot_patch
):
    al.plot.phase.interferometer_of_phase(
        interferometer=interferometer_7,
        kpc_per_arcsec=2.0,
        unit_label="kpc",
        plot_as_subplot=True,
        plot_visibilities=False,
        plot_uv_wavelengths=False,
        plot_primary_beam=True,
        visualize_path=phase_plotter_path,
        subplot_path=phase_plotter_path,
    )

    assert phase_plotter_path + "interferometer.png" in plot_patch.paths
    assert (
        phase_plotter_path + "interferometer/interferometer_visibilities.png"
        not in plot_patch.paths
    )
    assert (
        phase_plotter_path + "interferometer/interferometer_primary_beam.png"
        in plot_patch.paths
    )


def test__plot_ray_tracing_for_phase__dependent_on_input(
    tracer_x2_plane_7x7, sub_grid_7x7, mask_7x7, phase_plotter_path, plot_patch
):
    al.plot.phase.ray_tracing_of_phase(
        tracer=tracer_x2_plane_7x7,
        grid=sub_grid_7x7,
        during_analysis=False,
        plot_in_kpc=True,
        mask=mask_7x7,
        include_caustics=True,
        include_critical_curves=True,
        positions=None,
        plot_as_subplot=True,
        plot_all_at_end_png=False,
        plot_all_at_end_fits=False,
        plot_image=True,
        plot_source_plane=True,
        plot_convergence=False,
        plot_potential=True,
        plot_deflections=False,
        visualize_path=phase_plotter_path,
        subplot_path=phase_plotter_path,
    )

    assert phase_plotter_path + "tracer.png" in plot_patch.paths
    assert (
        phase_plotter_path + "ray_tracing/tracer_profile_image.png" in plot_patch.paths
    )
    assert (
        phase_plotter_path + "ray_tracing/tracer_source_plane.png" in plot_patch.paths
    )
    assert (
        phase_plotter_path + "ray_tracing/tracer_convergence.png"
        not in plot_patch.paths
    )
    assert phase_plotter_path + "ray_tracing/tracer_potential.png" in plot_patch.paths
    assert (
        phase_plotter_path + "ray_tracing/tracer_deflections_y.png"
        not in plot_patch.paths
    )
    assert (
        phase_plotter_path + "ray_tracing/tracer_deflections_x.png"
        not in plot_patch.paths
    )


def test__imaging_fit_for_phase__source_and_lens__depedent_on_input(
    masked_imaging_fit_x2_plane_7x7, phase_plotter_path, plot_patch
):
    al.plot.phase.imaging_fit_of_phase(
        fit=masked_imaging_fit_x2_plane_7x7,
        during_analysis=False,
        include_mask=True,
        include_positions=True,
        plot_in_kpc=True,
        include_critical_curves=True,
        include_caustics=True,
        include_image_plane_pix=True,
        plot_all_at_end_png=False,
        plot_all_at_end_fits=False,
        plot_fit_as_subplot=True,
        plot_fit_of_planes_as_subplot=False,
        plot_inversion_as_subplot=False,
        plot_image=True,
        plot_noise_map=False,
        plot_signal_to_noise_map=False,
        plot_model_image=True,
        plot_residual_map=False,
        plot_normalized_residual_map=True,
        plot_chi_squared_map=True,
        plot_inversion_residual_map=False,
        plot_inversion_normalized_residual_map=False,
        plot_inversion_chi_squared_map=False,
        plot_inversion_regularization_weights=False,
        plot_subtracted_images_of_planes=True,
        plot_model_images_of_planes=False,
        plot_plane_images_of_planes=True,
        visualize_path=phase_plotter_path,
        subplot_path=phase_plotter_path,
    )

    assert phase_plotter_path + "fit.png" in plot_patch.paths
    assert phase_plotter_path + "fit/fit_image.png" in plot_patch.paths
    assert phase_plotter_path + "fit/fit_noise_map.png" not in plot_patch.paths
    assert (
        phase_plotter_path + "fit/fit_signal_to_noise_map.png" not in plot_patch.paths
    )
    assert phase_plotter_path + "fit/fit_model_image.png" in plot_patch.paths
    assert phase_plotter_path + "fit/fit_residual_map.png" not in plot_patch.paths
    assert (
        phase_plotter_path + "fit/fit_normalized_residual_map.png" in plot_patch.paths
    )
    assert phase_plotter_path + "fit/fit_chi_squared_map.png" in plot_patch.paths
    assert (
        phase_plotter_path + "fit/fit_subtracted_image_of_plane_0.png"
        in plot_patch.paths
    )
    assert (
        phase_plotter_path + "fit/fit_subtracted_image_of_plane_1.png"
        in plot_patch.paths
    )
    assert (
        phase_plotter_path + "fit/fit_model_image_of_plane_0.png"
        not in plot_patch.paths
    )
    assert (
        phase_plotter_path + "fit/fit_model_image_of_plane_1.png"
        not in plot_patch.paths
    )
    assert phase_plotter_path + "fit/fit_plane_image_of_plane_0.png" in plot_patch.paths
    assert phase_plotter_path + "fit/fit_plane_image_of_plane_1.png" in plot_patch.paths


def test__interferometer_fit_for_phase__source_and_lens__depedent_on_input(
    masked_interferometer_fit_x2_plane_7x7, phase_plotter_path, plot_patch
):
    al.plot.phase.interferometer_fit_of_phase(
        fit=masked_interferometer_fit_x2_plane_7x7,
        during_analysis=False,
        include_mask=True,
        include_positions=True,
        plot_in_kpc=True,
        include_critical_curves=True,
        include_caustics=True,
        include_image_plane_pix=True,
        plot_all_at_end_png=False,
        plot_all_at_end_fits=False,
        plot_fit_as_subplot=True,
        plot_inversion_as_subplot=False,
        plot_visibilities=True,
        plot_noise_map=False,
        plot_signal_to_noise_map=False,
        plot_model_visibilities=True,
        plot_residual_map=False,
        plot_normalized_residual_map=True,
        plot_chi_squared_map=True,
        plot_inversion_residual_map=False,
        plot_inversion_normalized_residual_map=False,
        plot_inversion_chi_squared_map=False,
        plot_inversion_regularization_weights=False,
        visualize_path=phase_plotter_path,
        subplot_path=phase_plotter_path,
    )

    assert phase_plotter_path + "fit.png" in plot_patch.paths
    assert phase_plotter_path + "fit/fit_visibilities.png" in plot_patch.paths
    assert phase_plotter_path + "fit/fit_noise_map.png" not in plot_patch.paths
    assert (
        phase_plotter_path + "fit/fit_signal_to_noise_map.png" not in plot_patch.paths
    )
    assert phase_plotter_path + "fit/fit_model_visibilities.png" in plot_patch.paths
    assert phase_plotter_path + "fit/fit_residual_map.png" not in plot_patch.paths
    assert (
        phase_plotter_path + "fit/fit_normalized_residual_map.png" in plot_patch.paths
    )
    assert phase_plotter_path + "fit/fit_chi_squared_map.png" in plot_patch.paths


def test__hyper_images_for_phase__source_and_lens__depedent_on_input(
    hyper_model_image_7x7, mask_7x7, phase_plotter_path, plot_patch
):
    al.plot.phase.plot_hyper_images_for_phase(
        hyper_model_image=hyper_model_image_7x7,
        hyper_galaxy_image_path_dict=None,
        kpc_per_arcsec=2.0,
        unit_label="arcsec",
        mask=mask_7x7,
        plot_hyper_model_image=True,
        plot_hyper_galaxy_images=False,
        visualize_path=phase_plotter_path,
    )

    assert (
        phase_plotter_path + "hyper_galaxies/hyper_model_image.png" in plot_patch.paths
    )
    assert (
        phase_plotter_path + "hyper_galaxies/hyper_galaxy_images.png"
        not in plot_patch.paths
    )



from autoarray.plotters.array_plotters import plot_array as array
from autoarray.plotters.grid_plotters import plot_grid as grid
from autoarray.plotters.line_yx_plotters import plot_line as line
from autoarray.plotters.quantity_radii_plotters import (
    plot_quantity_as_function_of_radius as quantity_as_function_of_radius,
)
from autoarray.plotters import imaging_plotters as imaging
from autoarray.plotters import interferometer_plotters as interferometer
from autoarray.plotters import fit_imaging_plotters as fit_imaging
from autoarray.plotters import mapper_plotters as mapper
from autoarray.plotters import inversion_plotters as inversion

from autoastro.plotters import profile_plotters as profile
from autoastro.plotters import galaxy_plotters as galaxy
from autoastro.plotters import fit_galaxy_plotters as fit_galaxy

from autolens.plotters import plane_plotters as plane
from autolens.plotters import ray_tracing_plotters as tracer
from autolens.plotters.fit_imaging_plotters import fit_imaging_plotters as fit_imaging
from autolens.plotters.fit_interferometer_plotters import (
    fit_interferometer_plotters as fit_interferometer,
)
from autolens.plotters import hyper_plotters as hyper
from autolens.plotters import phase_plotters as phase



import autofit as af
import autolens as al
from test_autolens.integration.tests.imaging import runner

test_type = "grid_search"
test_name = "multinest_grid_subhalo__parallel"
data_type = "lens_mass__source_smooth"
data_resolution = "lsst"


def make_pipeline(name, phase_folders, optimizer_class=af.MultiNest):
    class GridPhase(af.as_grid_search(al.PhaseImaging, parallel=True)):
        @property
        def grid_priors(self):
            return [
                self.model.galaxies.subhalo.mass.centre_0,
                self.model.galaxies.subhalo.mass.centre_1,
            ]

        def customize_priors(self, results):

            ### Lens Subhalo, Adjust priors to physical masses (10^6 - 10^10) and concentrations (6-24)

            self.galaxies.subhalo.mass.kappa_s = af.UniformPrior(
                lower_limit=0.0005, upper_limit=0.2
            )
            self.galaxies.subhalo.mass.centre_0 = af.UniformPrior(
                lower_limit=-2.0, upper_limit=2.0
            )
            self.galaxies.subhalo.mass.centre_1 = af.UniformPrior(
                lower_limit=-2.0, upper_limit=2.0
            )

    phase1 = GridPhase(
        phase_name="phase_1",
        phase_folders=phase_folders,
        galaxies=dict(
            lens=al.GalaxyModel(redshift=0.5, mass=al.mp.EllipticalIsothermal),
            subhalo=al.GalaxyModel(
                redshift=0.5, mass=al.mp.SphericalTruncatedNFWChallenge
            ),
            source=al.GalaxyModel(redshift=1.0, light=al.lp.EllipticalSersic),
        ),
        optimizer_class=optimizer_class,
        number_of_steps=2,
    )

    phase1.optimizer.const_efficiency_mode = True

    return al.PipelineDataset(name, phase1)


if __name__ == "__main__":
    import sys

    runner.run(sys.modules[__name__])



from test_autolens.integration.tests.imaging.lens__source import (
    lens_mass__source_x2,
    lens_light_mass__source__hyper_bg,
    lens_light_mass__source__hyper,
    lens_mass__source_x2__hyper,
    lens_light_mass__source,
)
from test_autolens.integration.tests.imaging.runner import run_a_mock


class TestCase:
    def _test_lens_light_mass__source(self):
        run_a_mock(lens_light_mass__source)

    def _test_lens_light_mass__source__hyper(self):
        run_a_mock(lens_light_mass__source__hyper)

    def _test_lens_light_mass__source__hyper_bg(self):
        run_a_mock(lens_light_mass__source__hyper_bg)

    def _test_lens_mass__source_x2(self):
        run_a_mock(lens_mass__source_x2)

    def _test_lens_mass__source_x2__hyper(self):
        run_a_mock(lens_mass__source_x2__hyper)



from test_autoarray.mock.mock_inversion import MockMapper
from test_autoastro.mock.mock_galaxy import MockHyperGalaxy


class MockTracer(object):
    def __init__(
        self,
        unblurred_image_1d,
        blurring_image_1d,
        has_light_profile,
        has_pixelization,
        has_hyper_galaxy,
        has_grid_mappers=False,
    ):

        self.unblurred_image_1d = unblurred_image_1d
        self.blurring_image_1d = blurring_image_1d
        self.has_light_profile = has_light_profile
        self.has_pixelization = has_pixelization
        self.has_hyper_galaxy = has_hyper_galaxy
        self.has_grid_mappers = has_grid_mappers

    @property
    def all_planes(self):
        return []

    @property
    def mappers_of_planes(self):
        return [MockMapper()]

    @property
    def regularization_of_planes(self):
        return [MockMapper()]

    @property
    def hyper_galaxies(self):
        return [MockHyperGalaxy(), MockHyperGalaxy()]



from autolens.fit.plotters import masked_imaging_fit_plotters
from test import simulate_util

# In this tutorial, we'll introduce a new pixelization, called an adaptive-pixelization. This pixelization doesn't use
# uniform grid of rectangular pixels, but instead uses ir'Voronoi' pixels. So, why would we want to do that?
# Lets take another look at the rectangular grid, and think about its weakness.

# Lets quickly remind ourselves of the image, and the 3.0" circular mask we'll use to mask it.
imaging = simulate_util.load_test_imaging(
    data_type="lens_light_dev_vaucouleurs", data_resolution="lsst"
)
mask = al.mask.circular(
    shape=imaging.shape, pixel_scales=imaging.pixel_scales, radius=3.0
)

# The lines of code below do everything we're used to, that is, setup an image and its grid, mask it, trace it
# via a tracer, setup the rectangular mapper, etc.
lens_galaxy = al.Galaxy(
    bulge=al.EllipticalDevVaucouleurs(
        centre=(0.0, 0.0), axis_ratio=0.9, phi=45.0, intensity=0.1, effective_radius=1.0
    )
)

masked_imaging = al.masked.imaging(imaging=imaging, mask=mask)

tracer = al.Tracer.from_galaxies(galaxies=[lens_galaxy])
fit = al.fit(masked_dataset=masked_imaging, tracer=tracer)
al.plot.fit_imaging.subplot(fit=fit, include_mask=True)



import autofit as af
import matplotlib

backend = af.conf.get_matplotlib_backend()
matplotlib.use(backend)

from matplotlib import pyplot as plt

import autoarray as aa
from autoarray.util import plotter_util


def subplot_of_hyper_galaxy(
    hyper_galaxy_image_sub,
    contribution_map_sub,
    noise_map_sub,
    hyper_noise_map_sub,
    chi_squared_map_sub,
    hyper_chi_squared_map_sub,
    mask=None,
    unit_label="arcsec",
    kpc_per_arcsec=None,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="hyper_galaxies",
):

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=6
    )

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    plt.subplot(rows, columns, 1)

    hyper_galaxy_image(
        hyper_galaxy_image=hyper_galaxy_image_sub,
        mask=mask,
        as_subplot=True,
        unit_label=unit_label,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )

    plt.subplot(rows, columns, 2)

    aa.plot.array(
        array=noise_map_sub,
        mask=mask,
        as_subplot=True,
        unit_label=unit_label,
        unit_conversion_factor=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )

    plt.subplot(rows, columns, 3)

    hyper_noise_map(
        hyper_noise_map=hyper_noise_map_sub,
        mask=mask,
        as_subplot=True,
        unit_label=unit_label,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )

    plt.subplot(rows, columns, 4)

    contribution_map(
        contribution_map=contribution_map_sub,
        mask=mask,
        as_subplot=True,
        unit_label=unit_label,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        position_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )

    plt.subplot(rows, columns, 5)

    chi_squared_map(
        chi_squared_map=chi_squared_map_sub,
        mask=mask,
        as_subplot=True,
        unit_label=unit_label,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )

    plt.subplot(rows, columns, 6)

    hyper_chi_squared_map(
        hyper_chi_squared_map=hyper_chi_squared_map_sub,
        mask=mask,
        as_subplot=True,
        unit_label=unit_label,
        kpc_per_arcsec=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        mask_pointsize=mask_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    plt.close()


def subplot_of_hyper_galaxy_images(
    hyper_galaxy_image_path_dict,
    mask,
    include_mask=True,
    unit_label="arcsec",
    kpc_per_arcsec=None,
    figsize=None,
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    titlesize=10,
    xlabelsize=10,
    ylabelsize=10,
    xyticksize=10,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_filename="hyper_galaxy_images",
    output_format="show",
):

    rows, columns, figsize_tool = plotter_util.get_subplot_rows_columns_figsize(
        number_subplots=len(hyper_galaxy_image_path_dict)
    )

    if not include_mask:
        mask = False

    if figsize is None:
        figsize = figsize_tool

    plt.figure(figsize=figsize)

    hyper_index = 0

    for path, hyper_galaxy_image_sub in hyper_galaxy_image_path_dict.items():

        hyper_index += 1

        plt.subplot(rows, columns, hyper_index)

        hyper_galaxy_image(
            hyper_galaxy_image=hyper_galaxy_image_sub,
            mask=mask,
            as_subplot=True,
            unit_label=unit_label,
            kpc_per_arcsec=kpc_per_arcsec,
            figsize=figsize,
            aspect=aspect,
            cmap=cmap,
            norm=norm,
            norm_min=norm_min,
            norm_max=norm_max,
            linthresh=linthresh,
            linscale=linscale,
            cb_ticksize=cb_ticksize,
            cb_fraction=cb_fraction,
            cb_pad=cb_pad,
            cb_tick_values=cb_tick_values,
            cb_tick_labels=cb_tick_labels,
            title="Hyper Galaxy = " + path[0],
            titlesize=titlesize,
            xlabelsize=xlabelsize,
            ylabelsize=ylabelsize,
            xyticksize=xyticksize,
            mask_pointsize=mask_pointsize,
            position_pointsize=position_pointsize,
            output_path=output_path,
            output_format=output_format,
            output_filename=output_filename,
        )

    plotter_util.output_subplot_array(
        output_path=output_path,
        output_filename=output_filename,
        output_format=output_format,
    )

    plt.close()


def hyper_model_image(
    hyper_model_image,
    mask=None,
    positions=None,
    image_plane_pix_grid=None,
    as_subplot=False,
    unit_label="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Hyper Model Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    grid_pointsize=1,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="hyper_model_image",
):
    """Plot the image of a hyper_galaxies model image.

    Set *autolens.datas.arrays.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    hyper_model_image : datas.imaging.datas.Imaging
        The hyper_galaxies model image.
    include_origin : True
        If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """

    aa.plot.array(
        array=hyper_model_image,
        mask=mask,
        grid=image_plane_pix_grid,
        points=positions,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def hyper_galaxy_image(
    hyper_galaxy_image,
    mask=None,
    positions=None,
    image_plane_pix_grid=None,
    as_subplot=False,
    unit_label="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Hyper Model Image",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    grid_pointsize=1,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="hyper_galaxy_image",
):
    """Plot the image of a hyper_galaxies galaxy image.

    Set *autolens.datas.arrays.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    hyper_galaxy_image : datas.imaging.datas.Imaging
        The hyper_galaxies galaxy image.
    include_origin : True
        If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """

    aa.plot.array(
        array=hyper_galaxy_image,
        mask=mask,
        grid=image_plane_pix_grid,
        points=positions,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def contribution_map(
    contribution_map,
    mask=None,
    positions=None,
    image_plane_pix_grid=None,
    as_subplot=False,
    unit_label="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Contribution Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    grid_pointsize=1,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="contribution_map",
):
    """Plot the image of a hyper_galaxies galaxy image.

    Set *autolens.datas.arrays.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    contribution_map : datas.imaging.datas.Imaging
        The hyper_galaxies galaxy image.
    include_origin : True
        If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """

    aa.plot.array(
        array=contribution_map,
        mask=mask,
        grid=image_plane_pix_grid,
        points=positions,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def hyper_noise_map(
    hyper_noise_map,
    mask=None,
    positions=None,
    image_plane_pix_grid=None,
    as_subplot=False,
    unit_label="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Hyper Noise-Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    grid_pointsize=1,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="hyper_noise_map",
):
    """Plot the image of a hyper_galaxies galaxy image.

    Set *autolens.datas.arrays.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    hyper_noise_map : datas.imaging.datas.Imaging
        The hyper_galaxies galaxy image.
    include_origin : True
        If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """

    aa.plot.array(
        array=hyper_noise_map,
        mask=mask,
        grid=image_plane_pix_grid,
        points=positions,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def chi_squared_map(
    chi_squared_map,
    mask=None,
    positions=None,
    image_plane_pix_grid=None,
    as_subplot=False,
    unit_label="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Chi-Squared Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    grid_pointsize=1,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="chi_squared_map",
):
    """Plot the image of a hyper_galaxies galaxy image.

    Set *autolens.datas.arrays.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    chi_squared_map : datas.imaging.datas.Imaging
        The hyper_galaxies galaxy image.
    include_origin : True
        If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """

    aa.plot.array(
        array=chi_squared_map,
        mask=mask,
        grid=image_plane_pix_grid,
        points=positions,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )


def hyper_chi_squared_map(
    hyper_chi_squared_map,
    mask=None,
    positions=None,
    image_plane_pix_grid=None,
    as_subplot=False,
    unit_label="arcsec",
    kpc_per_arcsec=None,
    figsize=(7, 7),
    aspect="square",
    cmap="jet",
    norm="linear",
    norm_min=None,
    norm_max=None,
    linthresh=0.05,
    linscale=0.01,
    cb_ticksize=10,
    cb_fraction=0.047,
    cb_pad=0.01,
    cb_tick_values=None,
    cb_tick_labels=None,
    title="Hyper Chi-Squared Map",
    titlesize=16,
    xlabelsize=16,
    ylabelsize=16,
    xyticksize=16,
    grid_pointsize=1,
    mask_pointsize=10,
    position_pointsize=10,
    output_path=None,
    output_format="show",
    output_filename="hyper_chi_squared_map",
):
    """Plot the image of a hyper_galaxies galaxy image.

    Set *autolens.datas.arrays.plotters.array_plotters* for a description of all input parameters not described below.

    Parameters
    -----------
    hyper_chi_squared_map : datas.imaging.datas.Imaging
        The hyper_galaxies galaxy image.
    include_origin : True
        If true, the origin of the datas's coordinate system is plotted as a 'x'.
    """

    aa.plot.array(
        array=hyper_chi_squared_map,
        mask=mask,
        grid=image_plane_pix_grid,
        points=positions,
        as_subplot=as_subplot,
        unit_label=unit_label,
        unit_conversion_factor=kpc_per_arcsec,
        figsize=figsize,
        aspect=aspect,
        cmap=cmap,
        norm=norm,
        norm_min=norm_min,
        norm_max=norm_max,
        linthresh=linthresh,
        linscale=linscale,
        cb_ticksize=cb_ticksize,
        cb_fraction=cb_fraction,
        cb_pad=cb_pad,
        cb_tick_values=cb_tick_values,
        cb_tick_labels=cb_tick_labels,
        title=title,
        titlesize=titlesize,
        xlabelsize=xlabelsize,
        ylabelsize=ylabelsize,
        xyticksize=xyticksize,
        grid_pointsize=grid_pointsize,
        mask_pointsize=mask_pointsize,
        point_pointsize=position_pointsize,
        output_path=output_path,
        output_format=output_format,
        output_filename=output_filename,
    )



import autofit as af


def pipeline_tag_from_pipeline_settings(
    hyper_galaxies=False,
    hyper_image_sky=False,
    hyper_background_noise=False,
    include_shear=False,
    fix_lens_light=False,
    pixelization=None,
    regularization=None,
    align_bulge_disk_centre=False,
    align_bulge_disk_axis_ratio=False,
    align_bulge_disk_phi=False,
    disk_as_sersic=False,
    align_light_dark_centre=False,
    align_bulge_dark_centre=False,
):

    hyper_tag = hyper_tag_from_hyper_settings(
        hyper_galaxies=hyper_galaxies,
        hyper_image_sky=hyper_image_sky,
        hyper_background_noise=hyper_background_noise,
    )

    include_shear_tag = include_shear_tag_from_include_shear(
        include_shear=include_shear
    )

    fix_lens_light_tag = fix_lens_light_tag_from_fix_lens_light(
        fix_lens_light=fix_lens_light
    )

    pixelization_tag = pixelization_tag_from_pixelization(pixelization=pixelization)

    regularization_tag = regularization_tag_from_regularization(
        regularization=regularization
    )

    bulge_disk_tag = bulge_disk_tag_from_align_bulge_disks(
        align_bulge_disk_centre=align_bulge_disk_centre,
        align_bulge_disk_axis_ratio=align_bulge_disk_axis_ratio,
        align_bulge_disk_phi=align_bulge_disk_phi,
    )

    disk_as_sersic_tag = disk_as_sersic_tag_from_disk_as_sersic(
        disk_as_sersic=disk_as_sersic
    )

    align_light_dark_centre_tag = align_light_dark_centre_tag_from_align_light_dark_centre(
        align_light_dark_centre=align_light_dark_centre
    )

    align_bulge_dark_centre_tag = align_bulge_dark_centre_tag_from_align_bulge_dark_centre(
        align_bulge_dark_centre=align_bulge_dark_centre
    )

    return (
        "pipeline_tag"
        + hyper_tag
        + include_shear_tag
        + fix_lens_light_tag
        + pixelization_tag
        + regularization_tag
        + bulge_disk_tag
        + disk_as_sersic_tag
        + align_light_dark_centre_tag
        + align_bulge_dark_centre_tag
    )


def hyper_galaxies_tag_from_hyper_galaxies(hyper_galaxies):
    """Generate a tag for if hyper_galaxies-galaxies are used in a hyper_galaxies pipeline to customize phase names.

    This changes the phase name 'pipeline_name__' as follows:

    fix_lens_light = False -> pipeline_name__
    fix_lens_light = True -> pipeline_name___hyper_galaxies
    """
    if not hyper_galaxies:
        return ""
    elif hyper_galaxies:
        return "_galaxies"


def hyper_image_sky_tag_from_hyper_image_sky(hyper_image_sky):
    """Generate a tag for if the sky-background is hyper as a hyper_galaxies-parameter in a hyper_galaxies pipeline to
    customize phase names.

    This changes the phase name 'pipeline_name__' as follows:

    fix_lens_light = False -> pipeline_name__
    fix_lens_light = True -> pipeline_name___hyper_bg_sky
    """
    if not hyper_image_sky:
        return ""
    elif hyper_image_sky:
        return "_bg_sky"


def hyper_background_noise_tag_from_hyper_background_noise(hyper_background_noise):
    """Generate a tag for if the background noise is hyper as a hyper_galaxies-parameter in a hyper_galaxies pipeline to
    customize phase names.

    This changes the phase name 'pipeline_name__' as follows:

    fix_lens_light = False -> pipeline_name__
    fix_lens_light = True -> pipeline_name___hyper_bg_noise
    """
    if not hyper_background_noise:
        return ""
    elif hyper_background_noise:
        return "_bg_noise"


def hyper_tag_from_hyper_settings(
    hyper_galaxies, hyper_image_sky, hyper_background_noise
):

    if not any([hyper_galaxies, hyper_image_sky, hyper_background_noise]):
        return ""

    hyper_galaxies_tag = hyper_galaxies_tag_from_hyper_galaxies(
        hyper_galaxies=hyper_galaxies
    )

    hyper_image_sky_tag = hyper_image_sky_tag_from_hyper_image_sky(
        hyper_image_sky=hyper_image_sky
    )

    hyper_background_noise_tag = hyper_background_noise_tag_from_hyper_background_noise(
        hyper_background_noise=hyper_background_noise
    )

    return (
        "__hyper"
        + hyper_galaxies_tag
        + hyper_image_sky_tag
        + hyper_background_noise_tag
    )


def include_shear_tag_from_include_shear(include_shear):
    """Generate a tag for if an external shear is included in the mass model of the pipeline and / or phase are fixed
    to a previous estimate, or varied during he analysis, to customize phase names.

    This changes the phase name 'pipeline_name__' as follows:

    fix_lens_light = False -> pipeline_name__
    fix_lens_light = True -> pipeline_name___with_shear
    """
    if not include_shear:
        return ""
    elif include_shear:
        return "__with_shear"


def fix_lens_light_tag_from_fix_lens_light(fix_lens_light):
    """Generate a tag for if the lens light of the pipeline and / or phase are fixed to a previous estimate, or varied \
     during he analysis, to customize phase names.

    This changes the phase name 'pipeline_name__' as follows:

    fix_lens_light = False -> pipeline_name__
    fix_lens_light = True -> pipeline_name___fix_lens_light
    """
    if not fix_lens_light:
        return ""
    elif fix_lens_light:
        return "__fix_lens_light"


def pixelization_tag_from_pixelization(pixelization):

    if pixelization is None:
        return ""
    else:
        return "__pix_" + af.conf.instance.label.get(
            "tag", pixelization().__class__.__name__, str
        )


def regularization_tag_from_regularization(regularization):

    if regularization is None:
        return ""
    else:
        return "__reg_" + af.conf.instance.label.get(
            "tag", regularization().__class__.__name__, str
        )


def align_bulge_disk_centre_tag_from_align_bulge_disk_centre(align_bulge_disk_centre):
    """Generate a tag for if the bulge and disk of a bulge-disk system are aligned or not, to customize phase names \
    based on the bulge-disk model. This changee the phase name 'pipeline_name__' as follows:

    bd_align_centres = False -> pipeline_name__
    bd_align_centres = True -> pipeline_name___bd_align_centres
    """
    if not align_bulge_disk_centre:
        return ""
    elif align_bulge_disk_centre:
        return "__bd_align_centre"


def align_bulge_disk_axis_ratio_tag_from_align_bulge_disk_axis_ratio(
    align_bulge_disk_axis_ratio
):
    """Generate a tag for if the bulge and disk of a bulge-disk system are aligned or not, to customize phase names \
    based on the bulge-disk model. This changes the phase name 'pipeline_name__' as follows:

    bd_align_axis_ratio = False -> pipeline_name__
    bd_align_axis_ratio = True -> pipeline_name___bd_align_axis_ratio
    """
    if not align_bulge_disk_axis_ratio:
        return ""
    elif align_bulge_disk_axis_ratio:
        return "__bd_align_axis_ratio"


def align_bulge_disk_phi_tag_from_align_bulge_disk_phi(align_bulge_disk_phi):
    """Generate a tag for if the bulge and disk of a bulge-disk system are aligned or not, to customize phase names \
    based on the bulge-disk model. This changes the phase name 'pipeline_name__' as follows:

    bd_align_phi = False -> pipeline_name__
    bd_align_phi = True -> pipeline_name___bd_align_phi
    """
    if not align_bulge_disk_phi:
        return ""
    elif align_bulge_disk_phi:
        return "__bd_align_phi"


def bulge_disk_tag_from_align_bulge_disks(
    align_bulge_disk_centre, align_bulge_disk_axis_ratio, align_bulge_disk_phi
):
    """Generate a tag for the alignment of the geometry of the bulge and disk of a bulge-disk system, to customize \
    phase names based on the bulge-disk model. This adds together the bulge_disk tags generated in the 3 functions
    above
    """
    align_bulge_disk_centre_tag = align_bulge_disk_centre_tag_from_align_bulge_disk_centre(
        align_bulge_disk_centre=align_bulge_disk_centre
    )
    align_bulge_disk_axis_ratio_tag = align_bulge_disk_axis_ratio_tag_from_align_bulge_disk_axis_ratio(
        align_bulge_disk_axis_ratio=align_bulge_disk_axis_ratio
    )
    align_bulge_disk_phi_tag = align_bulge_disk_phi_tag_from_align_bulge_disk_phi(
        align_bulge_disk_phi=align_bulge_disk_phi
    )

    return (
        align_bulge_disk_centre_tag
        + align_bulge_disk_axis_ratio_tag
        + align_bulge_disk_phi_tag
    )


def disk_as_sersic_tag_from_disk_as_sersic(disk_as_sersic):
    """Generate a tag for if the disk component of a bulge-disk light profile fit of the pipeline is modeled as a \ 
    Sersic or the default profile of an Exponential.

    This changes the phase name 'pipeline_name__' as follows:

    disk_as_sersic = False -> pipeline_name__
    disk_as_sersic = True -> pipeline_name___disk_as_sersic
    """
    if not disk_as_sersic:
        return ""
    elif disk_as_sersic:
        return "__disk_sersic"


def align_light_dark_centre_tag_from_align_light_dark_centre(align_light_dark_centre):
    """Generate a tag for if the bulge and disk of a bulge-disk system are aligned or not, to customize phase names \
    based on the bulge-disk model. This changee the phase name 'pipeline_name__' as follows:

    bd_align_centres = False -> pipeline_name__
    bd_align_centres = True -> pipeline_name___bd_align_centres
    """
    if not align_light_dark_centre:
        return ""
    elif align_light_dark_centre:
        return "__light_dark_align_centre"


def align_bulge_dark_centre_tag_from_align_bulge_dark_centre(align_bulge_dark_centre):
    """Generate a tag for if the bulge and dark of a bulge-dark system are aligned or not, to customize phase names \
    based on the bulge-dark model. This changee the phase name 'pipeline_name__' as follows:

    bd_align_centres = False -> pipeline_name__
    bd_align_centres = True -> pipeline_name___bd_align_centres
    """
    if not align_bulge_dark_centre:
        return ""
    elif align_bulge_dark_centre:
        return "__bulge_dark_align_centre"
